/*
  Copyright (C) 2001 Jesper Hansen <jesperh@telia.com>.

  Written by  Romuald Bialy (aka MIS) <romek_b@o2.pl>

  Some ideas and optimisation method from Nikolai Vorontsov <nickviz@mail.be>

  This file is part of the yampp system.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <io.h>
#include <interrupt.h>
#include <progmem.h>

#include "atapi.h"
#include "delay.h"
#include "mem.h"
#include "uart.h"

#define ASM_OPT		// To speed up read procedures

static u08  SetAddress(u08 addr);
static u08  ReadBYTE(u08 addr);
static void WriteBYTE(u08 addr, u08 dat);
static void CheckBusy(void);
static u08  CheckDataReq(void);
extern bool bPlaying;

#define MINIBUFFERSIZE 64
u08 minibuffer[MINIBUFFERSIZE];

//----------------------------------------------------------------------------
// Init ATA
//----------------------------------------------------------------------------
void ATA_Init(void)
{
  	WriteBYTE(CMD_A | CMD_DEV_HEAD, DH_CHS0);	// Set Drive/Head register,
							// ie. select drive 0
	delay10us();
	CheckBusy();					// Wait for !BUSY
}


//----------------------------------------------------------------------------
// ATAPI Reset device
//----------------------------------------------------------------------------
void ATA_SW_Reset(void)
{
/*
	WriteBYTE(CTRL_A | CTRL_DEV_CTRL, CT_SRST | CT_nIEN | 8);
	delay10us();
	WriteBYTE(CTRL_A | CTRL_DEV_CTRL, CT_nIEN | 8);
	delay10us();
	CheckBusy();
*/
	delay10us();
	WriteBYTE(CMD_A | CMD_COMMAND, ATAPI_RESET);	// ATAPI Soft Reset command
	delay10us();
	CheckBusy();
	delay10(10000);
}

//----------------------------------------------------------------------------
// Clear Packet command buffer;
//----------------------------------------------------------------------------
void ATAPI_ClearBuf(void)
{
	register u08 i;
	for (i=0; i<12; i++)
		minibuffer[i] = 0;
}


//----------------------------------------------------------------------------
// Read one 2048 bytes sector in LBA mode from CD
// Returns 0 if no error detected
//----------------------------------------------------------------------------
u08 ATAPI_ReadLBA(u32 LBA, u08 *Buffer)
{
register u08 i,err;
register u08 errcount = 0;

#ifdef ASM_OPT
		register u08 j;
		asm volatile ("mov r26,r20");
		asm volatile ("mov r27,r21");
#endif
	union u32convert LBAC;
	LBAC.value = LBA;				// Prepare parameters...

	ATAPI_ClearBuf();
	minibuffer[0]=ATAPI_READ12;			// Read(12) Command
	minibuffer[2]=LBAC.bytes.byte4;
	minibuffer[3]=LBAC.bytes.byte3;
	minibuffer[4]=LBAC.bytes.byte2;
	minibuffer[5]=LBAC.bytes.byte1;
	minibuffer[9]=1;				// One sector (2048 bytes)

//	PRINT("LBA:"); UART_Printfu32(LBA); EOL();

CMDIN:	err = 0;
	while(ATAPI_Command() && (++err < 5));		// Issue ATAPI Packet command for READ(12)
	delay10(300);
	if (++errcount > 5)
	{
		ATA_SW_Reset();
		delay10(1000);
		return ERR_CMD;				// Too more errors > EXIT
	}

	if (CheckDataReq())				// If error
	{
		err = ReadBYTE(CMD_A | CMD_ERROR);
//		UART_Printfu08(err); UART_SendByte(32); UART_Printfu08(errcount); EOL();
		if (err == 0x20)
			return ERR_NOCD;		// Media not present > EXIT
		goto CMDIN;
	}

	if (ReadBYTE(CMD_A | CMD_CYL_HIGH) != 8)		// Check data size
		goto CMDIN;
	cli();

#ifdef ASM_OPT
		i = 0;
		do						// cycle sector by minibuffers 
		{

#ifdef YAMPP3USB
			SetAddress(CMD_A + CMD_DATA);			// setup addressing and chip selects
#endif
			outp(0x00, DDRA);			// port A as input
			outp(0x00, DDRC);			// port C as input
			outp(0xff, PORTA);			// activate Pullups
			outp(0xff, PORTC);			// activate Pullups
#ifndef YAMPP3USB
			SetAddress(CMD_A + CMD_DATA);		// setup addressing and chip selects
			cbi(MCUCR, SRE);			// Disable ExtRAM and fill
#endif								// the temporary buffer
			asm volatile ("ldi r30,lo8(minibuffer)");
			asm volatile ("ldi r31,hi8(minibuffer)");
			j = 0;
			do					// cycle minibuffer by words
			{
				cbi(PORTB, 1);			// set DIOR lo
				j += 2;				// nop replacement, allow pin change
				inp(PINA);			// read lo byte	
				asm volatile ("st Z+,r24");	// store to minibuf
				inp(PINC);			// read hi byte
				asm volatile ("st Z+,r24");	// store to minibuf
				sbi(PORTB, 1);			// set DIOR hi
			} while (j < MINIBUFFERSIZE);
			sbi(MCUCR, SRE);			// enable ExtRAM
			asm volatile ("ldi r30,lo8(minibuffer)");
			asm volatile ("ldi r31,hi8(minibuffer)");
			j = 0;
			do
			{
				asm volatile ("ld r24,Z+");
				asm volatile ("st X+,r24");
			} while (++j < MINIBUFFERSIZE);
		} while (++i < (BPS/MINIBUFFERSIZE));
#else

	u08 *miniPtr;
	for (i = 0; i < (BPS/MINIBUFFERSIZE); i++)	// Get size of data from CD-ROM
	{
#ifdef YAMPP3USB
		SetAddress(CMD_A + CMD_DATA);		// setup addressing and chip selects
#endif
		outp(0x00, DDRA);			// port A as input
		outp(0x00, DDRC);			// port C as input
		outp(0xff, PORTA);			// activate Pullups
		outp(0xff, PORTC);			// activate Pullups
#ifndef YAMPP3USB
		SetAddress(CMD_A + CMD_DATA); 		// setup addressing and chip selects
		cbi(MCUCR, SRE);			// Disable ExtRAM and store data into
#endif							//  the temporary buffer
		for (miniPtr = minibuffer; miniPtr < &minibuffer[MINIBUFFERSIZE];)
		{
			cbi(PORTB, 1);			// set DIOR lo
			asm volatile ("nop");		// allow pin change
			*miniPtr++ = inp(PINA);		// read lo byte
			*miniPtr++ = inp(PINC);		// read hi byte
			sbi(PORTB, 1);			// set DIOR hi
		}
		sbi(MCUCR, SRE);			// enable ExtRAM
		for (miniPtr = minibuffer; miniPtr < &minibuffer[MINIBUFFERSIZE];)
			*Buffer++ = *miniPtr++;
	} // for (i < BPS / MINIBUF

#endif // ASM_OPT

	ReadBYTE(CTRL_A | CTRL_ALT_STAT);
	sei();						// enable interrupt
	CheckBusy();
	return (ReadBYTE(CMD_A | CMD_STATUS) & SR_ERR);
}


//----------------------------------------------------------------------------
// Select address and CS signals
//
// addressing bits
// 35 DA0	A0	0x01	Address Line 0
// 33 DA1	A1	0x02	Address Line 1
// 36 DA2	A2	0x04	Address Line 2
//
// chip selects
// 37 CS0	A3 	0x08	Command Block Select
// 38 CS1	A4	0x10	Control Block Select
//
//
//----------------------------------------------------------------------------
u08 SetAddress(u08 addr)
{
#ifdef YAMPP3USB
	cbi(MCUCR, SRE);		// Disable ExtRAM
	outp(0xff, DDRA);
	outp(addr, PORTA);
	sbi(PORTE,PE1);
	cbi(PORTE,PE1);
	return 0;
#else
	return *(u08 *)(0xE000 + addr);
#endif
}


//----------------------------------------------------------------------------
// Read data BYTE from Drive
//----------------------------------------------------------------------------
u08 ReadBYTE(u08 addr)
{
register u08 tmp;
	SetAddress(addr);
#ifndef YAMPP3USB
	cbi(MCUCR, SRE);		// disable RAM
#endif
	outp(0x00, DDRA);		// port A as input
	outp(0x00, DDRC);		// port C as input
	outp(0xff, PORTA);		// activate Pullups
	asm volatile ("nop");		// allow pin change
	cbi(PORTB, 1);			// set DIOR lo
	asm volatile ("nop");		// allow pin change
	tmp = inp(PINA);		// read byte
	sbi(PORTB, 1);			// set DIOR hi
	sbi(MCUCR, SRE);		// enable RAM
  	return tmp;
}


//----------------------------------------------------------------------------
// Write data BYTE to Drive
//----------------------------------------------------------------------------
void WriteBYTE(u08 addr, u08 dat)
{
  	SetAddress(addr);
#ifndef YAMPP3USB
	cbi(MCUCR, SRE);		// disable RAM
	outp(0xff, DDRA);		// port A as output
#endif
	outp(0xff, DDRC);		// port C as output
	outp(dat, PORTA);		// write byte
	outp(0,   PORTC);
	asm volatile ("nop");		// allow pin change
	cbi(PORTB, 0);			// set DIOW lo
	asm volatile ("nop");		// allow pin change
	sbi(PORTB, 0);			// set DIOW hi
	sbi(MCUCR, SRE);		// enable RAM
}

//----------------------------------------------------------------------------
// Send the PACKET command to CD-ROM and 12 command data bytes from minibuffer
//----------------------------------------------------------------------------
u08 ATAPI_Command()
{
	register u08 i = 0;
	CheckBusy();
	WriteBYTE(CMD_A | CMD_FEATURES , 0);
	WriteBYTE(CMD_A | CMD_CYL_HIGH , 8);			// Max transfer length
	WriteBYTE(CMD_A | CMD_CYL_LOW ,  0);			// = 2048 bytes
	WriteBYTE( CMD_A | CMD_COMMAND, ATAPI_PACKET );		// Write command Packet
	delay10us();
	CheckBusy();
	cli();
  	SetAddress(CMD_A | CMD_DATA);
	outp(0xff, DDRC);					// port C as output
#ifndef YAMPP3USB
	outp(0xff, DDRA);					// port A as output
	cbi(MCUCR, SRE);					// disable RAM
#endif
	for(i=0; i<12;)
	{
		outp(minibuffer[i++], PORTA);			// write lo byte
		outp(minibuffer[i++], PORTC);			// write hi byte
		asm volatile ("nop");				// allow pin change
		cbi(PORTB, 0);					// set DIOW lo
		asm volatile ("nop");				// allow pin change
		sbi(PORTB, 0);					// set DIOW hi
	}
	sbi(MCUCR, SRE);					// enable RAM
	ReadBYTE(CTRL_A | CTRL_ALT_STAT);
	sei();
	CheckBusy();
	return (ReadBYTE(CMD_A | CMD_STATUS) & SR_ERR);		// Return command status
}


//----------------------------------------------------------------------------
// Check CD-ROM status
//----------------------------------------------------------------------------
u08 ATAPI_CheckStatus(void)
{
	register u08 err;
	ATAPI_ClearBuf();
	ATAPI_Command();

//	PRINT_p(PSTR("STAT : ")); UART_Printfu08(ReadBYTE(CMD_A | CMD_STATUS)); 
	err = ReadBYTE(CMD_A | CMD_ERROR) & 0x61;

//	PRINT_p(PSTR(" ERR : ")); UART_Printfu08(err); EOL();

	if (err == 0x60)
		return CD_STAT_MAN_CLOSE;
	if (err == 0x20)
		return CD_STAT_NODISK;
	if (err != 0)
		return CD_STAT_UNDEF_ERROR;
	return 0;
}

//----------------------------------------------------------------------------
// Slow down CD-ROM unit
//----------------------------------------------------------------------------
void SlowDown()
{
#ifdef READ_SPEED
	ATAPI_ClearBuf();
	minibuffer[0] = ATAPI_SLOWDOWN;
	minibuffer[2] = (READ_SPEED)>>8;
	minibuffer[3] = (u08)(READ_SPEED);
	ATAPI_Command();
#endif
}

//----------------------------------------------------------------------------
// Load or Unload CD
//----------------------------------------------------------------------------
void CD_Cmd(u08 Funct)
{
	ATAPI_ClearBuf();
	minibuffer[0] = ATAPI_START_STOP_UNIT;
	minibuffer[1] = 0x01;			// Immed = 1
	minibuffer[4] = Funct;
	ATAPI_Command();
}


//----------------------------------------------------------------------------
// Check ATA busy status
//----------------------------------------------------------------------------
void CheckBusy(void)
{
	u16 tout = 0;
	WDR;
	while (((ReadBYTE(CMD_A | CMD_STATUS) & SR_BUSY) == SR_BUSY) && (++tout < 20000)) delay10us();
}

//----------------------------------------------------------------------------
// Check ATA data request status
//----------------------------------------------------------------------------
u08 CheckDataReq(void)
{
	u16 timeout = 0;
	u16 maxtout = 50000 * (bPlaying == 0) + 3000;
	register u08 stat = 0;
	WDR;
	do
	{
		stat = ReadBYTE(CMD_A | CMD_STATUS);
		if (stat & SR_ERR) 
			return 1;
		if (++timeout > maxtout) 
			return 2;
		delay10(2);
	} while ((stat & (SR_BUSY | SR_DRQ)) != SR_DRQ);
	return 0;
}

