/*
  Copyright (C) 2001 Jesper Hansen <jesperh@telia.com>.

  Rewritten by:	Romuald Bialy (aka MIS) <romek_b@o2.pl>

  Some ideas from Nikolai Vorontsov <nickviz@mail.be>

  This file is part of the yampp system.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __MEM_H__
#define __MEM_H__

#include "types.h"
#include "Constants.h"

// 32 k External RAM

#ifdef YAMPP3USB
 #define EXTRAM_START	0x460
#else
 #define EXTRAM_START	0x260
#endif
#define EXTRAM_END		0x7FFF	

// 3 bytes _before_ LONG_NAME are also used!!!

#define LONG_NAME		((u08*)(EXTRAM_START + 0x00A0))	// length 256
#define DIR_NAME		((u08*)(EXTRAM_START + 0x01A0))	// length 256
#define CD_NAME			((u08*)(EXTRAM_START + 0x02A0))	// length 256
#define DIRS_TABLE		(EXTRAM_START + 0x03A0)		// length max 255 dirs * 6B = 0x0600
#define RANDOM_TAB		((u08*)(EXTRAM_START + 0x09A0))	// length max 255

#define DIR_CACHE		((u08*)0x1000)  		// length 2048


// free 0x1800 - 0x5fff

// This is the size of each buffer

#define BUFFER_SIZE		2048		// Don't change size in this version !!!
#define BUFFER1			0x6000
#define BUFFER2			(BUFFER1 + BUFFER_SIZE)
#define BUFFER1P		((u08*)BUFFER1)
#define BUFFER2P		((u08*)BUFFER2)

// eeprom positions
#define EEPROM_VOLUME		0x0010	/* u08 */
#define EEPROM_RANDOM		0x0011	/* u08 */
#define EEPROM_LOUDNESS		0x0012	/* u08 */
#define EEPROM_SONG_NUMBER	0x0013  /* u08 */
#define EEPROM_MAX_SONGS	0x0014  /* u08 */

struct DT_elem
{
	bool	dir_start;		// flag for dir skipping
	u08	song_num;
	u32	dir_block;
};

#define DirsTable ((struct DT_elem*)DIRS_TABLE)
#define MAX_DIRS	255

#endif // __MEM_H__
