/* ***********************************************************************
**
**  Copyright (C) 2002 	Romuald Bialy (MIS) <romek_b@o2.pl>.
**
**
**  Yampp-3/USB - Input control routines
**
**  File control.c
**
*************************************************************************
**
**   This file is part of the yampp system.
**
**  This program is free software; you can redistribute it and/or
**  modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2
**  of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software Foundation, 
**  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**
*************************************************************************
**
**  Revision History
**
**  when         what  who      why
**
**  2002-09-01   1.0   MIS      initial public release
**
*********************************************************************** */


#include <timer.h>
#include "yampp3_usb.h"


static u08 fn0[]  __attribute__ ((progmem)) = "PLAY";
static u08 fn1[]  __attribute__ ((progmem)) = "STOP";
static u08 fn2[]  __attribute__ ((progmem)) = "NEXT";
static u08 fn3[]  __attribute__ ((progmem)) = "PREV";
static u08 fn4[]  __attribute__ ((progmem)) = "NEXT PL";
static u08 fn5[]  __attribute__ ((progmem)) = "PREV PL";
static u08 fn6[]  __attribute__ ((progmem)) = "F.FWD";
static u08 fn7[]  __attribute__ ((progmem)) = "F.REW";
static u08 fn8[]  __attribute__ ((progmem)) = "M.BASS";
static u08 fn9[]  __attribute__ ((progmem)) = "RANDOM";
static u08 fn10[] __attribute__ ((progmem)) = "REPEAT";
static u08 fn11[] __attribute__ ((progmem)) = "TIME";
static u08 fn12[] __attribute__ ((progmem)) = "UP";
static u08 fn13[] __attribute__ ((progmem)) = "DOWN";
static u08 fn14[] __attribute__ ((progmem)) = "PLAYLIST";
static u08 fn15[] __attribute__ ((progmem)) = "MENU";

u08 *function_name[] = {(u08*)&fn0,(u08*)&fn1,(u08*)&fn2,(u08*)&fn3,(u08*)&fn4,(u08*)&fn5,
			(u08*)&fn6,(u08*)&fn7,(u08*)&fn8,(u08*)&fn9,(u08*)&fn10,(u08*)&fn11,
			(u08*)&fn12,(u08*)&fn13,(u08*)&fn14,(u08*)&fn15 };




#define IR_PORT		PORTE
#define IR_BIT		PE0
#define KBD_PIN		PIND
#define KBD_BIT		PD3

/****************************************************************************************
*			INFRARED REMOTE RECEPTION FUNCTIONS				*
****************************************************************************************/

u08 rec80_active(void)
{
	register u08 i = 50;

	while (i-- != 0)
		if (bit_is_clear(IR_PORT-2, IR_BIT))
			return 1;
	return 0;
}

static inline void timer0_src(u08 src)
{
	outb (src, TCCR0);
}

//---------------------------------------------------------------------------------------

/*
  if STD = 0
	The REC-80 format used by Panasonic is a space coded 48 bit code consisting 
	of a 32 bit group id, followed by a 16 bit commmand word.
	Leading this is a header consisting of a 10T signal and a 4T pause.
	All bits start with a pulse of length T. The length of the pause following
	indicates the bit value. A T pause for a 0-bit and a 3T pause for a 1-bit.

  if STD = 1
	The NEC format is a space coded 32 bit code consisting 
	of a 16 bit group id, followed by a 16 bit commmand word.
	Leading this is a header consisting of a 16T signal and a 8T pause.
	All bits start with a pulse of length T. The length of the pause following
	indicates the bit value. A T pause for a 0-bit and a 3T pause for a 1-bit.
	Modified REC80 format by Will Jenkins, wdj@cus.org.uk

  if STD = 2 or 3
	The SONY format is a pulse coded 15 or 12 bit code consisting 
	of a 6 bit device code, followed by a 6 or 9 bit commmand word.
	Leading this is a header consisting of a 4T signal pulse.
	All bits are composed of a pulse of either 1T or 2T in length followed by a
	1T pause. A 2T pulse indicates a 1 bit, and a 1T pulse is a 0 bit.

 if STD = 4
	The RC-5 format used by Philips is bi-phase coded 13 bit code consisting 
	of a 7 bit device code, followed by 6 bit commmand word. Highest two bits of
	device code is startbit and troggle bit. Troggle bit must be masked out 
	because	its change after any keypress.

*/

u16 get_rec80(u08 std)
{
		u08 i, T2, T4, time, tmp = 0;
		union u16convert code;

		code.value = 0;
		timer0_src(CK256);				//update every 32us
		timer0_start();
		loop_until_bit_is_set(IR_PORT-2, IR_BIT);	// skip leading signal


	if(std < 2)
	{	
		timer0_start();

		while (bit_is_set(IR_PORT-2, IR_BIT))
		{
			T2 = inp(TCNT0);
			if (T2 >= 140)				// max wait time
				return 0;
		}
	
		// measure time T

		timer0_start();
		loop_until_bit_is_set(IR_PORT-2, IR_BIT);
	
		T2 = inp(TCNT0);				// T is normally around 0E-10 hex = 15 -> 480 uS
		T2 = T2 * 2;
		// max time is 4T
		T4 = T2 * 2;
		
		for (i = 0; i < ((std == 0) ? 48 : 32); i++)
		{
			timer0_start();
			while(1)
			{
				time = inp(TCNT0);
				if (time > T4)
					return 0;
			
				// measure time on the lo flank
				if (bit_is_clear(IR_PORT-2, IR_BIT))
				{
					tmp <<= 1;
					if (time >= T2)
						tmp++;
					break;
				}
			}


			// save command data as we go
			if (std == 0)
			{
				if( i == 39)
					code.bytes.low = tmp;
				if( i == 47)
					code.bytes.high = tmp;
			}
			else
			{
				if( i == 15)
					code.bytes.high = tmp;
				if( i == 31)
					code.bytes.low = tmp;
			}

			// syncronize - wait for next hi flank
			loop_until_bit_is_set(IR_PORT - 2, IR_BIT);
		}
		return (code.value);
	}
	else if(std < 4)					// Format used by SONY remote controllers
	{
		if (inp(TCNT0) <= 60) 
			return 0;				// is it the leading Signal ?			

		// now looking for the Data-bits
		for(i=0; i < ((std == 3) ? 11 : 14); i++)
		{
			tmp = 0x01;
			while (bit_is_set(IR_PORT-2, IR_BIT))
			{
				T2 = inp(TCNT0);
				if (T2 >= 140)			// max wait time
					return 0;
			}
			timer0_start();
			while (bit_is_clear(IR_PORT-2, IR_BIT))
			{
				T2 = inp(TCNT0);
				if (T2 >= 140)			// max wait time
					return 0;
			}
			if (inp(TCNT0) >= 25)			// pulse longer than 1 ms? then logic "1"
				code.value += ((u16)tmp << i);	// the device code 
		}
		return (code.value);
	}

	else	// std=4 -> RC-5;

	{

		for(i=0; i<13; i++)
		{
			if(bit_is_clear(IR_PORT-2, IR_BIT) )
				T2 = 0;
			else
				T2 = 1;

			timer0_start();
			while(1)
			{
				time=inp(TCNT0);
				if(time > 0x21)
					return 0;

				if(bit_is_clear(IR_PORT-2, IR_BIT) && (T2==1) )
				{
					tmp <<= 1;
					tmp++;
					break;
				}
				else if(bit_is_set(IR_PORT-2, IR_BIT) && (T2==0) )
				{
					tmp <<= 1;
					break;
				}
			}

			//save address data
			if(i == 6)
			{
				code.bytes.high = (tmp & 0x5f);		// save address and cut troggle bit
				tmp=0;
			}

			//delay
			timer0_start();
			while(1)
			{
				time=inp(TCNT0);
				if(time > 0x21)
					break;
			}
		}
		code.bytes.low = tmp;
		return(code.value);
	}
}



/****************************************************************************************
*			KEYBOARD CONTROL FUNCTION					*
****************************************************************************************/

extern u08 nKeyTime;
u08 TableKbdShort[] = KBD_SHORT;	// Key definitions moved to "Constants.h" file !!!
u08 TableKbdLong[] = KBD_LONG;

event_e get_key_event(void)
{
	event_e event = EV_IDLE;
	static u08 keydown = 0;
	register u08 i;
	ramdisable();			// disable ExtRAM
	outp(0xff, PORTA-1);		// set port as output
	outp(0, PORTA);			// set all pins to 0
	asm volatile("nop");		// allow time for port to activate

	if (keydown)			// if a key was pressed earlier
	{
		// see if it has been released
		if (bit_is_set(KBD_PIN, KBD_BIT))			// check bit
		{
			// no key is down, return last key seen
			if (nKeyTime < 8 && nKeyTime > 0)		// max 0.8 sec for shortpress
				event = TableKbdShort[keydown];		// short keypress event
			keydown = 0;
		}
		else
			if ((nKeyTime & 0x7F) >= 8)			// long keypress ? 
			{
				event = TableKbdLong[keydown];		// long keypress event
				if (event == EV_UP || event == EV_DOWN || 
				    event == EV_FFWD  || event == EV_FREW)
					nKeyTime = 0x86;		// 0.2 sec repeat time for volume and fast forward
				else
					nKeyTime = 0x82;		// 0.6 sec repeat time
			}	


	}
	else
	{	
		if (bit_is_clear(KBD_PIN, KBD_BIT))			// check if a key is down
		{
			// a key is active, check which one
			for (i = 0; i < 8; i++)
			{
				outp(~(1 << i), PORTA);			// write bit mask
				asm volatile("nop");			// wait a while
				asm volatile("nop");
				if (bit_is_clear(KBD_PIN, KBD_BIT)) 	// this bit ?
				{
					keydown = i + 1;	// store the key
					nKeyTime = 0;		// clear keypress timer
					break;			// and exit
				}
			}
		}
	}

	ramenable();					// enable ExtRAM
	return ((event < EV_LASTEVENT) ? event : EV_IDLE);
}



/****************************************************************************************
*			MAIN PLAYER CONTROL FUNSTIONS					*
****************************************************************************************/

extern u08 ir_standard;

void control_handler(void)
{
	static event_e OldEvent = EV_IDLE;
	u08 c,i;
	event_e event = EV_IDLE;
		
	static u08 uart_chars[]="gGnPsSfFbretudlmN#!#";			// Tokens tabele for UART control

	if (uart_haschar())						// Check UART event
	{		
		c = uart_getchar();					// get char

		for(i=0; i < sizeof(uart_chars); i++)
		{
			if(c == uart_chars[i])				// check tokens tabele
			{
				event = i+1;				// found
				break;
			}
		}
	}

/*
		switch ((c = uart_getchar())) 
	    {

			case 'u' : event = EV_UP; break;
			case 'd' : event = EV_DOWN; break;

			case 'N' : event = EV_NEUTRAL; break; 
	
			case 'p' : event = EV_PREV; break;
			case 'n' : event = EV_NEXT; break;
			
			case 'g' : event = EV_PLAY; break;
			case 'G' : event = EV_STOP; break;
			
			case 'l' : event = EV_PLAYLIST; break;
			case 'm' : event = EV_MENU; break;

			case 'f' : event = EV_FFWD; break;
			case 'F' : event = EV_FREW; break;
			
			case 's' : event = EV_NEXTPL; break;
			case 'S' : event = EV_PREVPL; break;

			case 'b' : event = EV_MBASS; break;
			case 'r' : event = EV_RANDOM; break;
			case 'e' : event = EV_REPEAT; break;
			case 't' : event = EV_TIME; break;

			case '!' : event = EV_REMOTE_CFG; break;

			default:   event = EV_IDLE; break;

	    }
	}

*/

	if ((ir_standard != 255) && rec80_active())			// Check IR event
	{
		static u16 Lcod;
		register u08 i;
		u16 code = get_rec80(ir_standard);
		if (!(Lcod == code && nKeyTime <= DOUBLE_TRAP))		// Trap for some IR transmitters that fast repeat
		{
			for (i = 0; i < (2 * EV_LASTEVENT); i+=2)
				if (eeprom_rw(EEPROM_RemCodes + i) == code)
				{
					OldEvent = i/2 + 1;	// Save last event
					goto RET_LAST;
				}

			if ((nKeyTime >= 2) && (OldEvent == EV_UP || OldEvent == EV_DOWN || 
						OldEvent == EV_FFWD  || OldEvent == EV_FREW))
									// Autorepeat feature on some keys
			{	
RET_LAST:			nKeyTime = 0;
				Lcod = code;
				event = OldEvent;
			}
		}
	}

	else
	{
		if (nKeyTime > 16)
			OldEvent = EV_IDLE;
	}

	if (event == EV_IDLE)
		event = get_key_event();				// Check keyboard

	if (event != EV_IDLE)
		set_event(event);					// Add event to queue
}

/****************************************************************************************
*			SETUP REMOTE CODES FUNSTIONS					*
****************************************************************************************/

static u16 ee_adr;

void get_rem_code(u08* name, u08 stand)
{
	u16 code = -1;
	event_e event;
	do
	{
		lcd_clrline(1);
		uart_EOL();
		both_progputs(name);
		both_putchar(':');

		while((event = get_event()) == EV_IDLE)
		{
			if (rec80_active())
			{
				if ((code = get_rec80(stand)))
				{
					both_putchar('\r');
					both_progputs(name);
					fprintf(lcd_putchar,": %04x",code);
					fprintf(uart_putchar,": %04x",code);
					eeprom_ww(ee_adr, code);
				}
			}
		}
		if (event == EV_STOP)
		{
			eeprom_ww(ee_adr, 0xffff);
			break;
		}
	} while(event != EV_PLAY);
	ee_adr += 2;
	return;
}

static u08 ir0[]  __attribute__ ((progmem)) = "REC80 ";
static u08 ir1[]  __attribute__ ((progmem)) = "NEC80 ";
static u08 ir2[]  __attribute__ ((progmem)) = "SONY15";
static u08 ir3[]  __attribute__ ((progmem)) = "SONY12";
static u08 ir4[]  __attribute__ ((progmem)) = "RC-5  ";

u08 *standard_name[] = { (u08*)&ir0, (u08*)&ir1, (u08*)&ir2, (u08*)&ir3, (u08*)&ir4 };

void setup_rem(void)
{
	u08 stand = 0;
	u08 i=4;
	ir_standard = 255;
	ee_adr = EEPROM_RemCodes;

	lcd_clrscr();
	puts("\nRemote setup:\n");
	both_progputs(PSTR("Standard: 0 ?"));
	while(1)					// Recognize IR standard
	{
		if (rec80_active())
		{
			lcd_gotoxy(10,0);
			if (get_rec80(stand))		// If != 0  - found standard
			{
					uart_EOL();
					both_progputs(standard_name[stand]);
					delayms(3000);
					break;
			}
			else
			{
				if(--i == 0)
				{
					stand = (stand+1) % STANDARDS_COUNT;
					i=4;
					both_putchar(stand+'0');
				}
//				while(rec80_active());
			}
		}
		if(get_event() == EV_STOP)
		{
			set_event(EV_STOP);
			return;
		}
	}
	
	lcd_clrscr();
	uart_EOL();
	both_progputs(PSTR("PLAY-Ok,STOP-Skip"));
	for(i=0; i < 16; i++)
		get_rem_code(function_name[i],stand);
	ir_standard = stand;
	while(ee_adr++ < EV_LASTEVENT * 2 + EEPROM_RemCodes)	// fill rest function with 0xffff
		eeprom_wb(ee_adr, 0xff);

	eeprom_wb(EEPROM_IRSTAND,ir_standard);
	lcd_clrscr();
	uart_EOL();
	both_progputs(PSTR("Done !"));
}

