/* ***********************************************************************
**
**  Copyright (C) 2002  Jesper Hansen <jesperh@telia.com> and 
**			Romuald Bialy (MIS) <romek_b@o2.pl>.
**
**
**  Yampp-3/USB - main file
**
**  File yampp3_usb.c
**
*************************************************************************
**
**   This file is part of the yampp system.
**
**  This program is free software; you can redistribute it and/or
**  modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2
**  of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software Foundation, 
**  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**
*************************************************************************
**
**  Revision History
**
**  when         what  who	why
**
**  2002-09-01   1.0   MIS      initial public release
**
*********************************************************************** */

#define YAMPP_FIRMWARE_VERSION_MAJOR	1
#define YAMPP_FIRMWARE_VERSION_MINOR	0

#define DISK_LAYOUT_VERSION_MAJOR	1
#define DISK_LAYOUT_VERSION_MINOR	0

static char firmware_scan_tag[] __attribute__ ((progmem)) = "yfwTAG";
static char model_txt[]         __attribute__ ((progmem)) = "yampp-3/USB";
static char date_txt[]          __attribute__ ((progmem)) = __DATE__" / "__TIME__;
static char version_txt[]       __attribute__ ((progmem)) = "1.00";


/*
	PIN assignements on the yampp3/USB board
	
	PA0-PA7		data bus 0..7 + Address 0..7
	
	PB0		T0	DIOW
	PB1		T1	DIOR
	PB2		AIN0	DREQ
	PB3		AIN1	BSYNC		
	PB4		SS	MP3
	PB5		MOSI	SO
	PB6		MISO 	SI
	PB7		SCK	SCK
	
	PC0-=C7		address 8..15	

	PD0		RxD	RS-232/RS-485 RX
	PD1		TxD	RS-232/RS-485 TX
	PD2		INT0	TX_ENABLE (alternatively KEY_INPUT2)
	PD3		INT1	KEY_INPUT
	PD4		T0	TXE
	PD5		T1	RXE
	PD6		WR	RD
	PD7		RD	WR	

	PE0		ICP/INT2	
	PE1		ALE	ALE
	PE2		OC1B	LCD_ENABLE


~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	The CPU is running with a 7.3728 MHz clock
*/


#include "yampp3_usb.h"

//------------------------------------------------------------------------------------------

// Amount of small steps in LCD play position bar
#define LCD_STEPS ((LCD_LINE_LENGTH - 6) * 5)

//------------------------------------------------------------------------------------------


#ifdef HEXDUMP
void HexDump( u08 *buffer, u16 len ) 
{
    u16 i,j,k = 0;

   
    setputchar(uart_putchar);

    printf("\r\n");
    for(i=0;i<len/16;i++,k+=16)
    {
    	printf("%04x ",k);
        for(j=0;j<16;j++)
        {
        	printf("%02x ",buffer[i*16+j]);
        }
        printf("    ");
        for(j=0;j<16;j++)
        {
            if (buffer[i*16+j] >= 0x20)
                putchar(buffer[i*16+j]);
            else
                putchar('.');
        }
        printf("\r\n");
    }

}

void sectordump(u32 sector)
{
	printf("SectorDump %x", sector & 0xffff); // TODO
	
	printf("\r\nAddr %x\r\n",&ATA_ReadBuffer); 

	ATA_Read(sector,1,buffer1);

    setputchar(uart_putchar);

	
	HexDump(buffer1, 512);
}

#endif

//------------------------------------------------------------------------------------------

void idle(void);
void next_song(u08 mode);

event_e menu_event;

u08 isPlaying;				// playing status
u08 usb_play;				// usb playing status
u08 bLoudness;				// loudness modificator
u08 bRandom;				// random modificator
u08 Repeat;				// repeat modificator
u08 info_flag;				// 0=enable display, 1=only scroll enabled, 2=disable update
u08 timemode;				// normal or remain time displaying modificator
u08 volume;				// volume level
u08 vol_decrease;			// modificator for fast forward and fast rewind functions
u16 BackStepPos;			// Actual position in backstep tabele
u08 OldSteps;				// For progressbar displaying

u08 *outptr;				// pointer for sending data to vs1001
u08 *buffer1;				// first play buffer begin pointer
u08 *buffer2;				// second play buffer begin pointer
u08 *buffer3;				// second play buffer end pointer
u08 *updbuf;				// pointer for buffer updating
u32 filesize;				// size of file to play (divided by 32)
u32 fileplayed;				// amount of played data (divided by 32)
u16 songtime;				// time of current song in seconds
u16 addtime;				// time calculation helper for fast forward and fast rewind functions
u16 bitrate;				// bitrate of current song

u08 *artistname = (u08*)ARTISTNAME_BUF;	// pointer to artistname buffer
u08 *titlename = (u08*)SONGNAME_BUF;	// pointer to songname buffer
u08 *scroll_line = (u08*)SCROLL_LINE;	// pointer to scroll line buffer
u08 scroll_length;			// size of scroll data
u08 scroll_pos;				// current position in scroll string

//---------------------------------------------------------------------------

void ATA_SW_Reset2(void)		// ATA Reset
{
	WriteBYTE(ATAPI_DevCtrl, 0x06);				// SRST and nIEN bits
	delay10();						// 10uS delay
	WriteBYTE(ATAPI_DevCtrl, 0x02);				// nIEN bit
	delay10();						// 10 uS delay
	while ( (ATA_WaitBusy() & SR_DRDY) != SR_DRDY ); 	// Wait for DRDY
  	ATA_ReadStat();
}

void ATA_Standby(u08 time)
{
	WriteBYTE(ATA_SectorCount, time);		// time * 5 second (0=disable)
	WriteBYTE(ATAPI_Cmd, 0xE3);			// set standby timer and execute IDLE
	ATA_WaitBusy();
}

//----------------------------------------------------------------------------

#define EVENT_QUEUE_SIZE	8
volatile event_e event_queue[EVENT_QUEUE_SIZE];
volatile u08 event_queue_head = 0;
volatile u08 event_queue_tail = 0;

void set_event(event_e e)
{
	event_queue[event_queue_head] = e;
	event_queue_head = (event_queue_head + 1) % EVENT_QUEUE_SIZE; 
}

event_e get_event(void)
{
	event_e tmp = EV_IDLE;
	
	//
	// let the system have it's time
	//
	idle();
	
	//
	// check queue
	//
	if (event_queue_head != event_queue_tail)
	{
		tmp = event_queue[event_queue_tail];
		event_queue_tail = (event_queue_tail + 1) % EVENT_QUEUE_SIZE; 
	}
	return tmp;
}

//--------------------------------------------------------------------------

// prescaler set to 34.72 uS
// 2880 ticks = 100 mS

#define TI1_H	(((u16)-2880) >> 8)
#define TI1_L	(((u16)-2880) & 0xff )

volatile u08 time_flag;			// counter for display refresh
volatile u08 scroll_time;

SIGNAL(SIG_OVERFLOW1)	//timer 1 overflow every 100 mS
{
	time_flag++;
	nKeyTime++;			// counter for keyboard and IR functions
	scroll_time++;
	outp(TI1_H, TCNT1H);		//reload timer 
	outp(TI1_L, TCNT1L);
}

//--------------------------------------------------------------------------

void send_sinewave_beeps(void)
{
	u08 i,j,buf[8];
	
	// sine on
	buf[0] = 0x53;	buf[1] = 0xEF;	buf[2] = 0x6E;	buf[3] = 0x30;
	// sine off
	buf[4] = 0x45;	buf[5] = 0x78;	buf[6] = 0x69;	buf[7] = 0x74;
	delayms(100);
	vs1001_nulls(4);
	for (j=0;j<3;j++)
	{
		for (i=0;i<4;i++)
			vs1001_send_data(buf[i]);
		vs1001_nulls(4);
		delayms(100);
		
		for (i=4;i<8;i++)
			vs1001_send_data(buf[i]);
		vs1001_nulls(4);
		delayms(100);
	}	
}

//--------------------------------------------------------------------------

void lcd_clrline(u08 line)		// clear single line of LCD
{
	register u08 i;
	lcd_gotoxy(0,line);
	for(i=0; i < LCD_LINE_LENGTH; i++)
		lcd_putchar(' ');
	lcd_gotoxy(0,line);		// goto begin of line
}

//--------------------------------------------------------------------------

void setvolume(u08 v)
{
	u08 i,j;
	vs1001_setvolume(v,v);
	eeprom_wb(EEPROM_VOLUME,v);
	if(!usb_play)
	{
		lcd_command(0x40);				// define of volume bar characters
		j=8;
		while(j--)
		{
			for (i = 0; i < 8; i++)
			{
				if (i >= j)
					lcd_data(0x1f);
				else
					lcd_data(0x00);
			}
		}

		lcd_clrline(LCD_LINES - 1);
		lcd_progputs(PSTR("Volume "));
		for (i=0; i < 8; i++)				// write volume bar
		{
			if(MIN_VOLUME - v >= i * (VOL_STEP*(NUM_VOL_STEPS/8)))
				lcd_putchar(i);
			else
				lcd_putchar(' ');
		}
		time_flag = 0;					// prevent display overvriting for 2 second
		info_flag = 1;					// allow scroll data
	}
}

void eeprom_ww(u16 address, u16 value)		// write word to eeprom
{
	union u16convert Data;
	Data.value = value;
	eeprom_wb(address, Data.bytes.low);
	eeprom_wb(address + 1, Data.bytes.high);
}


/************************************************************************
*				RANDOM STUFF				*
************************************************************************/

// Random stuff, maximum 65535 songs on the disk!

static u32 seed = 1;

// Init randomize table
void InitRnd(u16 MaxSongs)
{
u16 Temp;
	MaxSongs &= 0x7fff;						// limit for lower RAM usage
	for (Temp = 0; Temp < (MaxSongs >> 3); Temp++)
		RANDOM_TAB[Temp] = 0;					// one bit in table represent one song
	RANDOM_TAB[Temp] = (0xff << (MaxSongs % 8));			// last byte.
}

u16 randcalc(u16 max)							// assembler optimized randomizer function
{
//  return ((seed = seed * 1103515245L + 12345) % max);	// get new song number

	asm volatile ("push r24");
	asm volatile ("push r25");
	asm volatile ("lds r22,seed");
	asm volatile ("lds r23,(seed)+1");
	asm volatile ("lds r24,(seed)+2");
	asm volatile ("lds r25,(seed)+3");
	asm volatile ("ldi r18,lo8(1103515245)");
	asm volatile ("ldi r19,hi8(1103515245)");
	asm volatile ("ldi r20,hlo8(1103515245)");
	asm volatile ("ldi r21,hhi8(1103515245)");
	asm volatile ("call __mulsi3");
	asm volatile ("subi r22,lo8(-(12345))");
	asm volatile ("sbci r23,hi8(-(12345))");
	asm volatile ("sbci r24,hlo8(-(12345))");
	asm volatile ("sbci r25,hhi8(-(12345))");
	asm volatile ("sts seed,r22");
	asm volatile ("sts (seed)+1,r23");
	asm volatile ("sts (seed)+2,r24");
	asm volatile ("sts (seed)+3,r25");
	asm volatile ("pop r19");
	asm volatile ("pop r18");
	asm volatile ("clr r20");
	asm volatile ("clr r21");
	asm volatile ("call __udivmodsi4");
	asm volatile ("mov r25,r23");
	asm volatile ("mov r24,r22");
	return max;
}

u08 tabval(u16 num, u08 fn)					// check random tabele for song status
{								// and mark as played if necessary
	u16 index = num >> 3;
	u08 offset = 1 << (num % 8);
	if(fn)
		RANDOM_TAB[index] |= offset;			// if fn=1 mark as played
	return (RANDOM_TAB[index] & offset);			// return status
}

// new random procedure
u16 do_rand(u16 max)						// returned random value from 0 to max - 1
{
u16 num;
	max &= 0x7fff;						// limit for lower RAM usage
	for(num = 0; num < max; num++)
		if(tabval(num,0) == 0)
			break;
	if (num == max)
		num = -1;
	else
	{
		do
		{
//			WDR;
			num = randcalc(max);
		} while (tabval(num,0) != 0);			// check this song not be played piervously
		tabval(num,1);					// mark as just played
	}
	return num;
}


/************************************************************************
*				FORMAT NEW DISK				*
*************************************************************************/

u08 format_new_disk(void)
{
	YADL_ROOT_SECTOR *pBoot;					// Boot Sector struct
	IDENTIFY_DATA *pID;	
	u08 i,c;
	u32 sectors;
	
	// first get some drive data
	
	IdentifyDrive(buffer1);
	
	// flip and save the ModelNumber data
	pID = (IDENTIFY_DATA *) buffer1;
	strncpy(buffer1,(u08*)pID->ModelNumber,sizeof(pID->ModelNumber));
	for (i=0;i<sizeof(pID->ModelNumber);i+=2)
	{
		c = buffer1[i];
		buffer1[i] = buffer1[i+1];
		buffer1[i+1] = c;
	}
	buffer1[sizeof(pID->ModelNumber)] = 0;				// terminate entry
	
	// get max sectors (disk size)
	sectors = pID->UserAddressableSectors;
	
	// clear buffer
	pBoot = (YADL_ROOT_SECTOR *) buffer2;				// use temp buffer for room
	memset(pBoot,0,512);

// section 1
	strncpy(pBoot->id,"YADL",4);

	pBoot->disk_size = sectors;
	
	pBoot->unmount_state = YADL_STATE_NEW;
	pBoot->media_type = YADL_MEDIA_DISK;

	pBoot->disk_version_major = DISK_LAYOUT_VERSION_MAJOR;
	pBoot->disk_version_minor = DISK_LAYOUT_VERSION_MINOR;

	strncpy_P(pBoot->model,model_txt, YADL_LEN_MODEL);		// yampp model name

	// copy the ModelNumber result from Identify Drive
	strncpy(pBoot->disk_info, buffer1, YADL_LEN_DISK_INFO);	

// section 2
	
	strncpy_P(pBoot->build_timestamp, date_txt, YADL_LEN_BUILD_TIMESTAMP);
	pBoot->firmware_version_major = YAMPP_FIRMWARE_VERSION_MAJOR;
	pBoot->firmware_version_minor = YAMPP_FIRMWARE_VERSION_MINOR;
	
// section 3	
// section 4
// section 5

	// write the sector
	// and return result
	
	return (ATA_Write(0,1,(u08*) pBoot));
}


/****************************************************************************************/


u32 songbase_start;
u16 songbase_qty;			// number of songs
u32 fat_start;				// first fat sector
u32 sector_cluster_offset;		// add to cluster -> sector conversions
u32 playlist_start;			// first playlist sector
u16 playlist_qty;			// number of playlists


u32 MyCluster;				// current cluster number
u32 MySector;				// current sector number
u16 SectorsLd;				// sectors in cluster counter

/****************************************************************************************/


//
// return the next FAT in a chain
//
u32 next_fat(u32 current_fat)
{
	u16 fat_offset;
	u32 fat_sector;
	static u32 fat_in_cache = -1;
	
	fat_sector = fat_start + current_fat / (SECTORSIZE / sizeof(CLUSTER));
	fat_offset = current_fat % (SECTORSIZE / sizeof(CLUSTER));
	
	if (fat_sector != fat_in_cache)
	{
		while(ATA_Read(fat_sector,1,(u08*)FAT_CACHE))
		    	ATA_SW_Reset2();
		fat_in_cache = fat_sector;
	}
	if(BackStepPos < MAX_REV_STEPS)
		Rev_Table[++BackStepPos] = ((CLUSTER*)FAT_CACHE)[fat_offset];
	return ((CLUSTER*)FAT_CACHE)[fat_offset];
}

void init_load_sectors(u32 cluster)
{
	MyCluster = cluster;
	MySector = sector_cluster_offset + YADL_SECTORS_PER_CLUSTER * MyCluster;
	SectorsLd = 0;
}


//
// load a bufferfull of data
//
void load_sectors(u08 *buffer)
{
	while(ATA_Read( MySector, BUFFERSIZE / SECTORSIZE, buffer))
	    	ATA_SW_Reset2();
	MySector += BUFFERSIZE / SECTORSIZE;
	SectorsLd += BUFFERSIZE / SECTORSIZE;
	if (SectorsLd == YADL_SECTORS_PER_CLUSTER)
		init_load_sectors(next_fat(MyCluster));
}

/****************************************************************************************/

static char firmver_txt[] __attribute__ ((progmem)) = "Firmware version";

void init_stuff(void)
{
	u08 update = 0;
	YADL_ROOT_SECTOR *pRoot;				// Boot Sector struct
	pRoot = (YADL_ROOT_SECTOR *) buffer1;			// Allocate room from the temp buffer
	while(ATA_Read(0,1,buffer1))				// read root sector
		ATA_SW_Reset2();

	if(pRoot->id[0] != 'Y' || pRoot->id[1] != 'A' || 
	   pRoot->id[2] != 'D' || pRoot->id[3] != 'L')		// YADL signature not found
	{
		event_e event = EV_IDLE;
		lcd_clrscr();
		both_progputs(PSTR(" Found not YADL\n or new disk !!\n"));
#if (LCD_LINES == 2)
		delayms(5000);
		lcd_clrscr();
#endif
		both_progputs(PSTR("Press PLAY to\nformat the disk."));
		uart_EOL();
		while((event = get_event()) != EV_PLAY);	// wait for PLAY key
		lcd_clrscr();
		both_progputs(PSTR("All data be lost\nAre you sure ?\n"));

		while((event = get_event()) == EV_PLAY);	// wait for release key

		while((event = get_event()) != EV_PLAY);	// wait for accept by PLAY key

		if(format_new_disk() == 0)			// format the disk
		{
#if (LCD_LINES == 2)
			lcd_clrscr();
#endif
			both_progputs(PSTR("Formating OK."));
		}
		else
		{
#if (LCD_LINES == 2)
			lcd_clrscr();
#endif
			both_progputs(PSTR("Format ERROR !"));
		}
		wdt_enable(0x07);
		while(1);					// reset player by watchdog
	}

	if(pRoot->firmware_version_major != YAMPP_FIRMWARE_VERSION_MAJOR)
	{
		pRoot->firmware_version_major = YAMPP_FIRMWARE_VERSION_MAJOR;
		update = 1;
	}
	if(pRoot->firmware_version_minor != YAMPP_FIRMWARE_VERSION_MINOR)
	{
		pRoot->firmware_version_minor = YAMPP_FIRMWARE_VERSION_MINOR;
		update = 1;
	}
	if(strcmp_P(pRoot->build_timestamp, date_txt))
	{
		strcpy_P(pRoot->build_timestamp, date_txt);
		update = 1;
	}
	if (update)
	{
		while(ATA_Write(0,1,(u08*) pRoot))
		    	ATA_SW_Reset2();
		lcd_clrscr();
		both_progputs(firmver_txt);
		both_progputs(PSTR("\ninfo updated.\n"));
//		both_progputs(PSTR("Firmware version\ninfo updated.\n"));
		delayms(4000);
	}

	sector_cluster_offset = pRoot->sector_cluster_offset;
	songbase_start 	= pRoot->songbase_start;			// SECTOR !!
	songbase_qty	= pRoot->songbase_qty;
	playlist_start 	= sector_cluster_offset + YADL_SECTORS_PER_CLUSTER * pRoot->playlist_start;	// SECTOR !!!
	playlist_qty 	= pRoot->playlist_qty;		
	fat_start 	= pRoot->fat_start;
}


/****************************************************************************************/

u16 curPlaylist = 0;
u16 curIndex = 0;
u16 selected_Playlist;

/****************************************************************************************/


YADL_PLAY_LIST_ENTRY *get_pl_entry(u16 index)
{
	static u32 nLastSector = -1;
	u32 nCurSector;
	
	//
	// find (and load) correct playlist sector
	//

	nCurSector =  playlist_start + (index >> 3);	// 64 bytes per one playlist entry
	if (nCurSector != nLastSector)
	{
		while(ATA_Read(nCurSector,1,(u08*)PL_CACHE))
			ATA_SW_Reset2();
		nLastSector = nCurSector;	
	}
	return &((YADL_PLAY_LIST_ENTRY *)PL_CACHE)[index % 8];
}

/****************************************************************************************/

YADL_SONG_BASE_ENTRY *get_song_entry(u16 index)
{
	static u32 nLastSector = -1;
	u32 nCurSector;
	
	nCurSector = songbase_start + (index >> 2);
	if (nCurSector != nLastSector)
	{
		while(ATA_Read(nCurSector,1,(u08*)SE_CACHE))
			ATA_SW_Reset2();
		nLastSector = nCurSector;
	}
	return &((YADL_SONG_BASE_ENTRY *)SE_CACHE)[index % 4];
}

/****************************************************************************************/

u16 get_offset_entry(u16 index, u16 Playlist)
{
	static u32 nLastSector = -1;
	YADL_PLAY_LIST_ENTRY *pListEntry = get_pl_entry(Playlist);
	FOFFSET offset = pListEntry->offset;

	// first compute desired start sector
	u32 nCurSector = playlist_start + (offset / SECTORSIZE) + (index * sizeof(u16) / SECTORSIZE);

	offset = offset % SECTORSIZE;		 // cut offset to rest inside one sector

	if (nCurSector != nLastSector)
	{
		while(ATA_Read(nCurSector,2,(u08*)LO_CACHE))
			ATA_SW_Reset2();
		nLastSector = nCurSector;
	}

	u16 *pList = (u16*) ( (u08*)LO_CACHE + offset);  // again put offset here as it needs not to begin on sector boundary
	index = index % (SECTORSIZE/sizeof(u16));        // cut off rest of index that already is in other sectors
	return pList[index];
}

/****************************************************************************************/

void pl_disp(u16 index, u08 sel)
{
	YADL_PLAY_LIST_ENTRY *pEntry = get_pl_entry(index);
	if (index < playlist_qty)
	{
		lcd_puts(pEntry->name);					// print on LCD
		if(sel)							// if current index
			printf("\rPlaylist: %s     ",pEntry->name);	// print on terminal
	}
}

/****************************************************************************************/

void se_disp(u16 index, u08 sel)
{
	YADL_SONG_BASE_ENTRY *pEntry = get_song_entry(index);
	if (index < songbase_qty)
	{
		lcd_puts(pEntry->artist);
		lcd_putchar('\n');
		lcd_putchar((sel) ? '~' : ' ');
		lcd_puts(pEntry->title);			// print title
		if (sel)
			printf("\rSong: %s / %s     ",pEntry->artist,pEntry->title);
	}
}

/****************************************************************************************/

void lo_disp(u16 index, u08 sel)
{
	se_disp(get_offset_entry(index,selected_Playlist),sel);
}

/****************************************************************************************/

void menu_disp(u16 index, u08 sel)
{
	extern u08 *function_name[];
	lcd_progputs(function_name[index]);		// print on LCD
}

/****************************************************************************************/

//
//
//
u16 browse_list(u16 index, u16 min, u16 max, void (*disp)(u16, u08))
{
	u08 i,j = 0;
	event_e event;
	u08 k = (*disp == lo_disp) ? LCD_LINES/2 : LCD_LINES;
	while (1)
	{
		lcd_clrscr();
		for (i=0; i < k; i++)
		{
			idle();
			lcd_gotoxy(0,i * (1+(*disp == lo_disp)));
			lcd_putchar((i==j) ? '~' : ' ');
			disp(index+i,(i==j));
		}

		do 
		{
			time_flag = 0;
			event = get_event();
			switch (event)
			{
				case EV_NEXT:
				case EV_FFWD:
				case EV_UP:
					if ( index+j > min )
					{
						if(j)
							 j--;
						else
							index--;
					}
					break;
				case EV_PREV:
				case EV_FREW:
				case EV_DOWN:
					if ( index+j+1 < max )
					{
						if(j < k - 1)
							j++;
						else
							index++;
					}
					break;
				case EV_PLAY:
					return index+j;

				case EV_STOP:
					return -1;
			}
		}
		while (event == EV_IDLE);
	}
}


/****************************************************************************************/


u08 browse_playlist(void)
{
	if (curPlaylist == -1)
		curPlaylist = 0;
	selected_Playlist = browse_list(curPlaylist, 0, playlist_qty, pl_disp);
	if (selected_Playlist != -1)
		return 0;
	return 1;
}



/****************************************************************************************/


u16 browse_songs(void)
{
	YADL_PLAY_LIST_ENTRY *pListEntry = get_pl_entry(selected_Playlist);
	u16 entry_qty = pListEntry->entry_qty;
	return (browse_list((curPlaylist == selected_Playlist) ? curIndex : 0, 0, entry_qty, lo_disp));
}


/****************************************************************************************/

u16 browse_menu(void)
{
	return (browse_list((menu_event) ? menu_event-1 : 2, 2, 12, menu_disp));
}


/****************************************************************************************/


void lcd_frame(void)					// show song info and progressbar frame
{
	if(isPlaying && !info_flag)
	{
		u08 i;
		lcd_command(0x48);			// start definition of CG-RAM from char code 1
	
		for (i = 0; i < 7; i++)
			lcd_data(0x1f);			// value for full bar char (1)
		lcd_data(0x00);				// cursor line

		lcd_data(0x1f);				// empty bar definition (2) (frame)
		for (i = 0; i < 5; i++)
			lcd_data(0x00);
		lcd_data(0x1f);
		lcd_data(0x00);				// cursor line

		lcd_data(0x1f);				// empty ended bar definition (3) (frame end)
		for (i = 0; i < 5; i++)
			lcd_data(0x01);
		lcd_data(0x1f);
		lcd_data(0x00);				// cursor line

#if (LCD_LINES == 4)
		lcd_clrscr();
		lcd_puts(scroll_line);			// playlist name
		lcd_putchar('\n');
		lcd_puts(artistname);
		lcd_putchar('\n');
		lcd_puts(titlename);
#endif
		lcd_gotoxy(0, LCD_LINES -1);		// progressbar position

		for (i=0; i< (LCD_STEPS / 5) - 1; i++)	// display progress bar outline
			lcd_putchar(2);	
		
		lcd_putchar(3);				// end of progressbar
		time_flag = 6;				// provide fast display refresh
		OldSteps = -1;
	}
}


//
// Run this as often as possible
// preferably every time around the main loop
//
void idle(void)
{
	usb_handler();
	if(usb_link == 0)					// usb link disables other control
		control_handler();
	//
	// time to update a buffer ?
	//
	if (updbuf && (!usb_play))
	{
		load_sectors(updbuf);
		updbuf = (u08 *) 0;				// mark as done
	}
#ifdef FAST_IDLE_DISK
	else
	{	
		if (isPlaying)
			ATA_Read(MySector, 1, (u08*) BBK_MEMORY);
	}
#endif	
	//	
	//
	//
	if (isPlaying)
	{
		// pump some data to the decoder
		
		while (bit_is_set(PINB, PB2) && (!updbuf))	// while DREQ is hi 
		{						// and buffer no need reload
			//send data on SPI bus
			vs1001_send32(outptr);
			outptr += 32;

			// check for buffer wrap
			if (outptr == buffer2)			// buffer 1 empty
			{
				updbuf = buffer1;		// buffer 1 need reload
				if (usb_play)
					usb_response(0);
			}
			else if (outptr == buffer3)		// buffer 2 empty
			{
				updbuf = buffer2;		// buffer 2 need reload
				if (usb_play)
					usb_response(0);
				outptr = buffer1;		// play from buffer 1
			}

			fileplayed++;

			// check if we're done

			if (fileplayed >= filesize)
			{
//				printf(" Done.\r\n");
				if (!usb_play)
				{
					isPlaying = 0;		// the end is near 
					next_song(0);
				}
				break;
			}
		}

		if (vol_decrease && (time_flag == 4))		// If fast forward volume downed
		{
			vs1001_setvolume(volume,volume);		// Restore volume
			vol_decrease = 0;
		}


#if (LCD_LINES == 2)
		if (scroll_time >= 5 && info_flag < 2)
		{
			//
			// scrolling artist and name

			u08 nTemp;
			lcd_gotoxy(0, 0);
			for (nTemp = 0; nTemp < LCD_LINE_LENGTH; nTemp++)
				lcd_putchar(scroll_line[(scroll_pos + nTemp) % scroll_length]);
			scroll_pos++;
			scroll_time = 0;
		}
#endif

		if (time_flag >= 5 && !info_flag)
		{
			//
			// display bar function

			static u08 data[] = { 0x10, 0x18, 0x1C, 0x1E, 0x1F };

			register u08 i, a, nSteps = (u08)((fileplayed * (LCD_STEPS-1)) / filesize);

			if(nSteps != OldSteps)					// progressbar need update ?
			{							// this check is for for CPU time saving
				a = nSteps / 5;					// # of full bars

				lcd_command(0x40);
				lcd_data(0x1F);					// full top line (frame outline)
				for (i = 0; i < 5; i++)				// value for current progress char (0)
				{
					lcd_data(data[nSteps % 5] | (a == (LCD_STEPS / 5) - 1));
				}
				lcd_data(0x1F);					// full bottom line (frame outline)
				lcd_data(0x00);					// cursor line

				lcd_gotoxy(0,LCD_LINES - 1);
	
				for (i = 0; i < a; i++)
					lcd_putchar(1);				// print full bars

				if (a <= (LCD_STEPS / 5) - 1)			// jman to fix display overwrite
					lcd_putchar(0);	
#ifdef UART_PROGRESSBAR
				puts("\r[");					// uart progressbar function
				for (i=0; i < LCD_STEPS - 2; i+=2)
					uart_putchar((i<nSteps) ? '*' : ' ');
				uart_putchar(']');
#endif
				OldSteps = nSteps;
			}

			lcd_gotoxy(LCD_LINE_LENGTH - 6, LCD_LINES - 1);	// time display position

			u16 wPlayTime = vs1001_read(VS1001_PLAYTIME) + addtime;

#ifdef UART_SONGTIME
			printf("%3u:%02u\b\b\b\b\b\b", wPlayTime / 60, wPlayTime % 60); // print time on UART
#endif

			if (timemode == 1)
			{
				wPlayTime = songtime - wPlayTime + 1;	// calculate time remain
				if (wPlayTime < 0)
					wPlayTime = 0;
			}
			fprintf(lcd_putchar,"%3u:%02u", wPlayTime / 60, wPlayTime % 60); // print time on LCD

			if (timemode == 1)				// add a '-' sign to time remain
			{
				lcd_gotoxy(LCD_LINE_LENGTH - 6 + (wPlayTime < 600), LCD_LINES - 1);
				lcd_putchar('-');
			}

			time_flag = 0;
		}

	}//if (isPlaying)

	if((time_flag > 30) && (info_flag))				// display overvrite timer ends
	{
		info_flag = 0;						// enable status updating
		lcd_frame();						// show progress bar and other data on 4 lines lcd
		menu_event = EV_IDLE;
	}
}


void start_playing(u16 st_offset)
{
	updbuf = 0;
	outptr = buffer1 + (st_offset & 0xffe0);
	isPlaying = 1;
	OldSteps = -1;
	vs1001_reset(bLoudness);
}


void create_scroll_data(void)
{
#if (LCD_LINES == 2)
	strcat(scroll_line, " -> ");			// playlist name separator
	strcat(scroll_line, artistname);
	strcat(scroll_line, " - ");			// add artist separator
	strcat(scroll_line, titlename);
	strcat(scroll_line, " >> ");			// add ' >> ' sign
	scroll_pos = 0;
	scroll_length = 0;				// Calculate scroll length
	u08* src = scroll_line;
	while (*src++)
		scroll_length++;
#endif
}


void play_song(void)
{
	if(!info_flag)
	{
		lcd_clrscr();
		lcd_progputs(PSTR("Loading..."));
	}

	YADL_PLAY_LIST_ENTRY *pListEntry = get_pl_entry(curPlaylist);
	strcpy(scroll_line, pListEntry->name);				// playlist name

	u16 curSong = get_offset_entry(curIndex,curPlaylist);
	YADL_SONG_BASE_ENTRY *pEntry = get_song_entry(curSong);

	if(pEntry->state != YADL_STATE_READY)				// check for valid song entry
	{
		lcd_gotoxy(0,0);
		both_progputs(PSTR("\nSongBase Error !"));
		delayms(1000);
		set_event(EV_STOP);
		return;
	}

	strcpy(artistname, pEntry->artist);
	strcpy(titlename, pEntry->title);
	songtime = pEntry->play_time;					// get playing time of current song
	u16 br = pEntry->bitrate;					// get bitrate
	
	create_scroll_data();						// create scroll line data for 2 lines lcd

	//
	// calc number of VS1001 blocks
	//
	u16 garbage = pEntry->start_offset+((br & (~OPTION_BR_ID1_TAG)) != 0) ? 128 : 0;
									// start_offset points to first valid MPEG
									// frame (after any header), 14 bit of
									// bitrate indicate that file have ID3 tag
									// and this may be excluded (last 128 bytes)
	filesize = (pEntry->length - garbage) / 32;			// calculation number of blocks
	fileplayed = 0;
	addtime = 0;							// for search time calculations

	//
	// preload buffers
	//
	init_load_sectors(pEntry->start);
	BackStepPos = 0;						// init cluster tabele counter
	Rev_Table[BackStepPos] = MyCluster;				// save first cluster number of song to be played

	u08 i;
	for (i=0; i < (pEntry->start_offset / BUFFERSIZE) + 1; i++)
	{
		load_sectors(buffer1);					// skip first pEntry->start_offset of full buffers
	}
	load_sectors(buffer2);

	u08 *brstr = (br > OPTION_BR_VBR) ? "VBR, average " : "";	// test VBR signature
	bitrate = (br & BITRATE_MASK);					// mask OPTIONS in bitrate

	printf("\r\nSong: %u ,  %s - %s\r\nPlaying Time %u:%02u ,  Bitrate %s%ukbps\r\n",
		 curIndex+1, artistname, titlename, songtime / 60, songtime % 60, brstr, bitrate);

	//
	// start playing the song
	//
	eeprom_ww(EEPROM_PLAYLIST,curPlaylist);			// save playlist number
	eeprom_ww(EEPROM_SONGPOS,curIndex);			// save song number within current playlist
	start_playing(pEntry->start_offset % BUFFERSIZE);					
	lcd_frame();						// display song info on 4 lines lcd and progress bar outline
}

void plist_info(YADL_PLAY_LIST_ENTRY * pListEntry)		// print info about playlist on terminal
{
	printf("\n\rPlaylist: %s ,  Playlist Time %u:%02u:%02u\r\n",
		pListEntry->name, pListEntry->length_hours, 
		pListEntry->length_minutes, pListEntry->length_seconds);
}

void next_playlist(void)
{
	YADL_PLAY_LIST_ENTRY *pListEntry = get_pl_entry(curPlaylist);

	if(pListEntry->next_playlist != -1)				// check for next playlist number
		curPlaylist = pListEntry->next_playlist;
	else
		curPlaylist++;

	if((curPlaylist == -1) || (curPlaylist >= playlist_qty))
		curPlaylist = 0;
	pListEntry = get_pl_entry(curPlaylist);
	InitRnd(pListEntry->entry_qty);					// init randomizer table
	plist_info(pListEntry);						// print pl info
	curIndex = -1;							// next song function change this index to 0
	next_song(1);
}


void next_song(u08 mode)
{
	YADL_PLAY_LIST_ENTRY *pListEntry = get_pl_entry(curPlaylist);

	if (Repeat == 3 && mode == 0)			// repeat one song
		goto GO_PLAY;

	if (bRandom)
		curIndex = do_rand(pListEntry->entry_qty);
	else
	{
		if (mode == 2)
			curIndex--;
		else
			curIndex++;
	}


	if (curIndex < (pListEntry->entry_qty))		// if End of playlist
		goto GO_PLAY;
	else
	{
		InitRnd(pListEntry->entry_qty);

		if(Repeat == 2 || mode == 2)		// repeat playlist
		{
			if (bRandom)
				curIndex = do_rand(pListEntry->entry_qty);
			else
				curIndex = 0;
			goto GO_PLAY;
		}

		if(Repeat == 0)
		{
			set_event(EV_STOP);		// normal play, stop after end of playlist
			return;
		}

		if(Repeat == 1)				// normal play, next playlist after end of playlist
		{
			next_playlist();
GO_PLAY:		play_song();
		}
	}
}



void disp_mod(u08 mod)					// display ON/OFF information
{
	lcd_progputs((mod) ? PSTR("N ") : PSTR("FF"));
	time_flag = 0;					// prevent display overvrite
	info_flag = 1;					// allow scroll
}


//----------------------------------------------------------------------------
// Main part
//----------------------------------------------------------------------------
int main(void) 
{
 
 //------------------------------
 // Initialize 
 //------------------------------

//////////////////////////////////////////////////////////////////
// B - Port
//////////////////////////////////////////////////////////////////

	sbi(PORTB, 0);	// DIOW- hi
	sbi(DDRB, 0);	// pin PB0 is output, DIOW- 
	
	sbi(PORTB, 1);	// DIOR- hi
	sbi(DDRB, 1);	// pin PB1 is output, DIOR- 
	
	cbi(DDRB, 2);	// pin PB2 is input, DEMAND
	sbi(PORTB, 2);	// activate pullup
	
	// PB3 and PB4 is used for BSYNC and MP3 signals to VS1001

	// PB5 (MOSI) is used for SPI
	// PB6 (MISO) is used for SPI
	// PB7 (SCK) is used for SPI
	

//////////////////////////////////////////////////////////////////
// D - Port
//////////////////////////////////////////////////////////////////

	// PD0 and PD1 are TX and RX signals

	sbi(PORTD,PD3);		// pin PD3 is keyboard input, enable pullup

	sbi(PORTD,PD4);		// PD4 is available but need a pullup to avoid noise
	
	sbi(PORTD,PD5);		// PD5
	
	// PD6 and PD7 are RD and WR signals
	// but we need to set their default states used
	// when the external memory interface is disabled
	// they should then both be high to avoid disturbing
	// RAM contents
	
	sbi(DDRD, 6);		// as outputs
	sbi(DDRD, 7);

	sbi(PORTD, 6);		// and both hi
	sbi(PORTD, 7);


	// PORTE need setup of ALE in idle mode

	cbi(DDRE, PE0);		// PE0 is IR input
	sbi(PORTE, PE0);	// activate pullup
	sbi(DDRE, PE1);		// set ALE pin as output
	cbi(PORTE, PE1);	// and set it low (inactive)

//////////////////////////////////////////////////////////////////
// USB
//////////////////////////////////////////////////////////////////

	usb_init();


//////////////////////////////////////////////////////////////////
// Timer 1
//////////////////////////////////////////////////////////////////

	//
	// setup timer1 for a 100mS periodic interrupt
	//	
	
	outp(0, TCCR1A);
	outp(4, TCCR1B);	// prescaler /256  tPeriod = 32 uS
	outp(TI1_H, TCNT1H);	// load counter value hi
	outp(TI1_L, TCNT1L);	// load counter value lo


 //----------------------------------------------------------------------------
 // 
 //----------------------------------------------------------------------------

	
    // enable external SRAM and wait states 

	outp((1<<SRE)|(1<<SRW10), MCUCR);
	delayms(10);
	lcd_init(0, LCD_FUNCTION_DEFAULT, (LCD_LINES<<8)+LCD_LINE_LENGTH); 		// does clrscr too

	uart_init(UART_BAUD_SELECT);		// init RS-232 link
	setputchar(uart_putchar);

#if (LCD_LINE_LENGTH == 16)
	both_progputs(PSTR(" Hi, MP3 world!\nyampp3/USB Alive\n"));
#else
	both_progputs(PSTR("   Hi, MP3 world!\n yampp-3/USB  Alive\n"));
#endif

	uart_init(UART_BAUD_SELECT);		// init RS-232 link

	// setup some buffers
	buffer1 = (u08 *) BUFFERSTART;			// 8 k
	buffer2 = (u08 *) buffer1 + BUFFERSIZE;		// 8 k
	buffer3 = (u08 *) buffer2 + BUFFERSIZE;		// end pointer

	// say hello
	_p_puts(firmver_txt);
	uart_putchar(':');
	_p_puts(version_txt);
	puts(" - ");
	_p_puts(date_txt);
	uart_EOL();

	// init VS1001
	vs1001_init_io();				// init VS1001
	vs1001_setcomp(F_VS1001);

	bLoudness = eeprom_rb(EEPROM_LOUDNESS);
	vs1001_reset(bLoudness);

	// Init and wait for drive
  	puts("Disk Init... ");
	ATA_SetDrive(0);
	ATA_SW_Reset2();
   	puts("Ready!\n");

	// init randomizer from timer
	seed = 1 + inp(TCNT1L);

	if ((volume = eeprom_rb(EEPROM_VOLUME)) == 255)
		volume = 12;		// -6 dB

	usb_play = 1;			// for prevent volume bar displaying
	setvolume(volume);
	usb_play = 0;

	ir_standard = eeprom_rb(EEPROM_IRSTAND) % STANDARDS_COUNT;	// get remote controller standard

	// Send three sinewave beeps to indicate we're starting up okay
#ifdef PWR_BEEPS
	send_sinewave_beeps();
#endif
	// enable timer1 interrupt
	sbi(TIMSK, TOIE1);	 	

	init_stuff();

	// check for pressing local "PLAY" key to enter remote IR setup procedure
	ramdisable();
	outp(0xff, DDRA);					// set port as output
	outp(0xff, PORTA);
	asm volatile("nop");				// allow time for port to activate
	if(bit_is_set(PIND, PD3))			// if not permanent short to GND
	{
		outp(0xfe, PORTA);			// set D0 pin to 0
		asm volatile("nop");			// allow time for port to activate
		if(bit_is_clear(PIND, PD3))		// if KEY 0 (PLAY) pressed
			setup_rem();			// go to remote IR setup procedure
	}
	ramenable();

	bRandom = eeprom_rb(EEPROM_RANDOM) % 2;
	Repeat = eeprom_rb(EEPROM_REPEAT) % 4;

	InitRnd(0x7fff);

	// load autoplay and position settings

	if(eeprom_rb(EEPROM_AUTOPLAY) == 1)
		set_event(EV_PLAY);
	else
		set_event(EV_STOP);

	curIndex = eeprom_rw(EEPROM_SONGPOS);
	curPlaylist = eeprom_rw(EEPROM_PLAYLIST);

	if (curPlaylist >= playlist_qty)
		curPlaylist = curIndex = 0;
/*
	WriteBYTE(ATA_SectorCount, 0x7f);	// set APM level
	WriteBYTE(ATAPI_Feat, 0x05);		// Enable APM subcommand
	WriteBYTE(ATAPI_Cmd, 0xEF);		// set features command
	ATA_WaitBusy();
*/
	ATA_Standby(HDD_STANDBY_TIME / 5);	// set automatic standby timer


///////////////////////////////////////////////////////////////////////////////
// start things rolling
///////////////////////////////////////////////////////////////////////////////

	event_e event = EV_IDLE;
	YADL_PLAY_LIST_ENTRY *pListEntry;
	static char usb_txt[] __attribute__ ((progmem)) = "Playing via USB";

	//
	// main loop
	//
  	while (1) 
  	{
		event = get_event();

		switch (event)
		{
			default:
			case EV_IDLE: 
				break;

			case EV_PLAYLIST:
				event = EV_IDLE;
				info_flag = 2;					// disable displaying
				if (browse_playlist())				// browse playlist
					goto BROWSE_BREAK;

				uart_EOL();
				u16 Index = browse_songs();			// browse songs within playlist
				uart_EOL();
				if (Index == -1)
				{						// browse terminate
BROWSE_BREAK:				pListEntry = get_pl_entry(curPlaylist);	// reload playlist cache
					info_flag = 0;				// enable displaying
					lcd_clrscr();
					lcd_frame();
					break;
				}
				curPlaylist = selected_Playlist;
				curIndex = Index;
				pListEntry = get_pl_entry(curPlaylist);		// 
				InitRnd(pListEntry->entry_qty);
				isPlaying = 0;
				
			case EV_PLAY:
				if (menu_event)
				{
					set_event(menu_event);
					time_flag = 0;
					continue;
				}

				if (isPlaying)						// Pause function
				{
					while(get_event() != EV_IDLE);			// Wait for release key
					ATA_Standby(0);					// disable automatic standby timer
					lcd_gotoxy(LCD_LINE_LENGTH - 6, LCD_LINES - 1);
					lcd_progputs(PSTR("Paused"));
					do 
					{
						control_handler();
					} while(event_queue_head == event_queue_tail);
					get_event();
					ATA_Standby(HDD_STANDBY_TIME / 5);		// enable automatic standby timer
					break;
				}
				else							// Play function
				{
					if(curPlaylist < playlist_qty && playlist_qty > 0)
					{
						pListEntry = get_pl_entry(curPlaylist);
						plist_info(pListEntry);
						if(curIndex >= pListEntry->entry_qty)
							curIndex = 0;
						info_flag = 0;
						play_song();
						eeprom_wb(EEPROM_AUTOPLAY,1);			// store autoplay flag
					}
					break;
				}
				
			case EV_PLAY_USB:
				uart_EOL();
				_p_puts(usb_txt);
				uart_EOL();
				strcpy_P(scroll_line,usb_txt);
				usb_play = 1;
				create_scroll_data();
				printf("Song: %s - %s\r\n",artistname, titlename);
				start_playing(0);
				lcd_frame();
				updbuf = buffer2;	
				break;

			case EV_STOP:
				isPlaying = 0;
				usb_play = 0;
				vs1001_reset(0);
				lcd_clrscr();
				lcd_progputs((u08*)&model_txt[0]);
#if (LCD_LINE_LENGTH == 16)
				lcd_progputs(PSTR(" STOP"));
#else
				lcd_progputs(PSTR(" - STOP"));
#endif
				eeprom_wb(EEPROM_AUTOPLAY,0);			// store autoplay flag
				if(usb_link)
				{
					ATA_Standby(0);
					both_progputs(PSTR("\nPC Link Active.\n"));
					usb_response(0);
				}
				break;				

			case EV_DOWN:						// volume down
				if (volume <= MIN_VOLUME-VOL_STEP) 
					volume += VOL_STEP;
				else
					volume = MIN_VOLUME;
				goto SETVOL;

			case EV_UP: 						// volume up
				if (volume >= VOL_STEP) 
					volume -= VOL_STEP;
				else
					volume = 0;
SETVOL:			setvolume(volume);
				break;

			case EV_NEXT:
				if (isPlaying)
					next_song(1);
				break;

			case EV_PREV:
				if (isPlaying)
					next_song(2);
				break;

			case EV_NEXTPL:
				if (isPlaying)
					next_playlist();
				break;

			case EV_PREVPL:
				if (isPlaying)
				{
					if(--curPlaylist == -1)
						curPlaylist = playlist_qty - 1;		// go to last playlist
					pListEntry = get_pl_entry(curPlaylist);		// update playlist data
					InitRnd(pListEntry->entry_qty);
					curIndex = 0;					// first song from playlist
					plist_info(pListEntry);				// print pl info
					play_song();
				}
				break;

			case EV_FFWD:
				if (isPlaying && !usb_play)
				{
					MyCluster = next_fat(MyCluster);	// get next cluster position
					init_load_sectors(MyCluster);		// skip to next cluster

						// calculate new size of played data
SEARCHGO:				fileplayed = BackStepPos * (YADL_CLUSTERSIZE / 32);

					// calculation of playing time - small hazard :-)
					addtime = (u16)(fileplayed*32 / (bitrate * 125)) - 
							vs1001_read(VS1001_PLAYTIME); 

					load_sectors(buffer1);
					outptr = buffer1;			// play from buffer 1
					updbuf = buffer2;			// and next update buffer 2
					vs1001_setvolume(volume+40,volume+40);	// volume down during search
					time_flag = 5;				// show information immedialy
					vol_decrease = 1;			// setting of volume downed flag
				}
				break;

			case EV_FREW:
				if (isPlaying && !usb_play && BackStepPos > 0)
				{
					MyCluster = Rev_Table[--BackStepPos];
					init_load_sectors(MyCluster);
					lcd_frame();
					goto SEARCHGO;
				}
				break;

			case EV_NEUTRAL: 
				volume = 12;		// -6 dB
				goto SETVOL;

			case EV_MBASS:
				bLoudness = !bLoudness;
				lcd_clrline(LCD_LINES - 1);				// clear status line
				lcd_progputs(PSTR("MegaBass O"));
				disp_mod(bLoudness);					// show "FF" or "N "
				vs1001_write(VS1001_MODE, 0x0080 * (bLoudness != 0)); // write bit SM_BASS to VS1001K
				eeprom_wb(EEPROM_LOUDNESS, bLoudness);		// store actual Mbass to eeprom
				break;

			case EV_RANDOM:
				bRandom = !bRandom;
				lcd_clrline(LCD_LINES - 1);		// clear status line
				lcd_progputs(PSTR("Random O"));
				disp_mod(bRandom);			// show "FF" or "N "
				eeprom_wb(EEPROM_RANDOM,bRandom);
				break;

			case EV_REPEAT:
				Repeat = (Repeat + 1) % 4;
				lcd_clrscr();
				lcd_progputs(PSTR("Play Mode: "));
				lcd_putchar(Repeat + '1');
				lcd_putchar('\n');
				time_flag = 0;				// prevent display overvrite
				info_flag = 2;
				switch(Repeat)
				{
					case 0:	lcd_progputs(PSTR("Single Playlist"));
						break;
					case 1:	lcd_progputs(PSTR("All Playlists"));
						break;
					case 2:	lcd_progputs(PSTR("Repeat Playlist"));
						break;
					case 3:	lcd_progputs(PSTR("Repeat One Song"));
						break;
				}
				eeprom_wb(EEPROM_REPEAT,Repeat);
				break;

			case EV_TIME:
				timemode = !timemode;			// change mode of time displaying (normal<->remain)
				if(menu_event)
				{
					lcd_clrline(LCD_LINES - 1);	// clear status line
					if(timemode)
						lcd_progputs(PSTR("Time: REMAIN"));
					else
						lcd_progputs(PSTR("Time: NORMAL"));
					info_flag = 1;
				}
				break;

			case EV_MENU:
				info_flag = 2;				// disable displaying
				menu_event = browse_menu() + 1;
				if(menu_event)
					set_event(menu_event);
				else
				{
					info_flag = 0;			// enable displaying
					menu_event = EV_IDLE;
					lcd_frame();
				}
				break;

			case EV_REMOTE_CFG:
				if(!isPlaying)
					setup_rem();			// go to setup remote codes procedure
				break;

		}
		event = EV_IDLE;

  	} // while (1)    
  	
}

